#!/usr/bin/env bash

# This script contains wrappers for Linux commands.
# Usage: $ source ./gpudb_bash_utils.sh

# ---------------------------------------------------------------------------
# Functions to initialize vars.

# All non-sourced scripts should call this with a < 16 char string
# prefaced with 'gpudb', e.g. "gpudb_COMPONENT"
# Set the '/proc/pid/comm' to identify this script, default is 'bash'.
THIS_SCRIPT_COMM=bash # global
function gpudb_set_proc_comm
{
    THIS_SCRIPT_COMM="$1"

    if ! printf "%s" "$THIS_SCRIPT_COMM" > "/proc/$$/comm" ; then
        echo_status "Script" "Error setting /proc/$$/comm to '$THIS_SCRIPT_COMM'"
        return 1
    fi
}

# All scripts that source this file should call this function in the beginning to set
# path information used by some functions in this file.
# Usage: gpudb_set_this_script_info "${BASH_SOURCE[0]}" "$@"
function gpudb_set_this_script_info
{
    local BASH_SOURCE_0="$1"
    shift

    # The directory of where this script was run from.
    THIS_SCRIPT_DIR="$( cd "$( dirname "${BASH_SOURCE_0}" )" && pwd )" # NOTE: you probably want THIS_SCRIPT_LINK_DIR
    THIS_SCRIPT_NAME="${BASH_SOURCE_0##*/}"
    THIS_SCRIPT_LINK=$(readlink -f "${BASH_SOURCE_0}") # filepath that a symlink resolves to
    THIS_SCRIPT_LINK_DIR="$(dirname "${THIS_SCRIPT_LINK}")" # the actual path of the file
    THIS_SCRIPT="$THIS_SCRIPT_DIR/$THIS_SCRIPT_NAME"    # filepath where this was run from
    THIS_SCRIPT_ARGS=("$@")
    #echo "THIS: '$THIS_SCRIPT_DIR' '$THIS_SCRIPT_NAME' '${THIS_SCRIPT_ARGS[@]}'"
}

# Our info, should be this filename, not caller even if sourced.
UTILS_THIS_SCRIPT_DIR="$( cd "$( dirname "${BASH_SOURCE[0]}" )" && pwd )"
UTILS_THIS_SCRIPT_NAME="${BASH_SOURCE[0]##*/}"
UTILS_THIS_SCRIPT_LINK=$(readlink -f "${BASH_SOURCE[0]}") # filepath that a symlink resolves to
UTILS_THIS_SCRIPT_LINK_DIR="$(dirname "$UTILS_THIS_SCRIPT_LINK")" # the actual directory the script resides in
UTILS_THIS_SCRIPT="$UTILS_THIS_SCRIPT_DIR/$UTILS_THIS_SCRIPT_NAME"    # filepath where this was run from
UTILS_THIS_SCRIPT_ARGS=("$@")

GPUDB_START_DATETIME=$(date '+%Y-%m-%d_%H.%M.%S')

# ---------------------------------------------------------------------------
# Call this function to check if running as USER_NAME and if not attempt to
# use 'su' or 'sudo' to re-run this script as that user with the same args.
# This function should normally only be called by 'root' or USER_NAME,
# but does allow the owner of a given file/dir to run it for alternate installs.
# Example usage:
# gpudb_set_this_script_info "${BASH_SOURCE[0]}" "$@"
# if ! switch_user_to "$GPUDB_USER" "/opt/gpudb/core/bin" "VAR_A,VAR_B"; then
#    exit 1
# fi
# If get here, then running as user with env vars VAR_A and VAR_B else the script has exited.

function switch_user_to
{
    local USER_NAME=$1
    local USER_FILE_OR_DIR=${2-} # Optional path to file/dir who's owner user is allowed to run this this script, without switching user.
    local EXTRA_ENV_VARS=${3-}   # Optional comma-separated list of env var names to pass to new shell, e.g. "A,B,C"

    local RET=0
    local CURRENT_USER=$(whoami)
    GPUDB_RUNNING_AS_WARN=${GPUDB_RUNNING_AS_WARN:-0}

    if id -u "$USER_NAME" > /dev/null; then
        # Check if we're already the correct user.
        if [[ "a$CURRENT_USER" == "a$USER_NAME" ]]; then
            return 0
        fi

        local FILE_DIR_OWNER="$USER_NAME"
        if [ -n "$USER_FILE_OR_DIR" ] ; then
            if ! FILE_DIR_OWNER=$(stat -c '%U' "$USER_FILE_OR_DIR") ; then
                echo "Error : Path '$USER_FILE_OR_DIR' does not exist, exiting."
                exit 1
            fi
        fi

        if [[ "a$FILE_DIR_OWNER" == "a$USER_NAME" ]]; then
            # Run this script again as the right user

            if [ $EUID -ne 0 ]; then
                # Non-root users may be asked for  password, this is their prompt.
                echo_status "Running as user" "$USER_NAME"
            fi

            local SU_OUTPUT=""
            # Try 'su' before 'sudo', which may not be installed or have config issues.
            SU_OUTPUT=$(su -l "$USER_NAME" -c "whoami") && RET=$? || RET=$?

            if [ $RET -eq 0 ]; then
                local SU_CMD_ARGS=()

                if [ -n "$EXTRA_ENV_VARS" ]; then
                    local SU_EXTRA_ENV_VAR_ARR=()
                    IFS=, read -ra SU_EXTRA_ENV_VAR_ARR <<< "$EXTRA_ENV_VARS"
                    for v in "${SU_EXTRA_ENV_VAR_ARR[@]}" ; do
                        SU_CMD_ARGS+=("$v=\"${!v}\"") # VAR='foo' VAR2='bar' cmd...
                    done
                fi
                # Run this script again as the right user, using '-' for login to prompt for password.
                SU_CMD_ARGS+=("$THIS_SCRIPT" "${THIS_SCRIPT_ARGS[@]}") # preserve array for quoting
                local SU_CMD=$(printf '%q ' "${SU_CMD_ARGS[@]}")
                # Check for --session-command though -h help to avoid needing to ask for password again.
                #if su - $USER_NAME --session-command whoami > /dev/null 2>&1; then
                if su --help | grep -qF -- --session-command > /dev/null ; then
                    # Using '--session-command' so that ctrl-c in sleep is trapped correctly.
                    exec su - "$USER_NAME" --session-command "$SU_CMD" && RET=$? || RET=$?
                else
                    # Ubuntu's su doesn't have --session-command, but this version's -c works for us.
                    exec su - "$USER_NAME" -c "$SU_CMD" && RET=$? || RET=$?
                fi

                exit $RET # Shouldn't reach here from exec
            fi

            # Try sudo before giving up
            local SUDO_OUTPUT=""
            SUDO_OUTPUT=$(sudo -u "$USER_NAME" "whoami" 2>&1) && RET=$? || RET=$?

            if [ $RET -eq 0 ]; then
                local SUDO_PRESERVE_ENV=""
                if [ -n "$EXTRA_ENV_VARS" ]; then
                    SUDO_PRESERVE_ENV="--preserve-env=$EXTRA_ENV_VARS"
                fi
                local SUDO_CMD_ARGS=("$THIS_SCRIPT" "${THIS_SCRIPT_ARGS[@]}") # preserve array for quoting
                local SUDO_CMD=$(printf '%q ' "${SUDO_CMD_ARGS[@]}")
                exec sudo -u "$USER_NAME" -H $SUDO_PRESERVE_ENV sh -c "$SUDO_CMD" && RET=$? || RET=$?
            else
                echo "Error attempting to 'su' or 'sudo' from user: '$CURRENT_USER' to user: '$USER_NAME', exiting."
                echo "- Command: 'su - $USER_NAME whoami' ; Output: '$SU_OUTPUT'"
                echo "- Command: 'sudo -u $USER_NAME whoami' ; Output: '$SUDO_OUTPUT'"
                exit 1
            fi

            exit $RET # Shouldn't reach here from exec

        elif [[ "a$FILE_DIR_OWNER" == "a$CURRENT_USER" ]]; then
            # For developer installs, we allow running as the owner of the gpudb dir.
            # This is harmless if the user owns all the dirs and they always run as this user.
            if [[ "a$GPUDB_RUNNING_AS_WARN" != "a1" ]]; then
                export GPUDB_RUNNING_AS_WARN=1
                echo "Warning : Running GPUdb as user: '$CURRENT_USER' since they own the '$USER_FILE_OR_DIR' directory."
            fi
            USER_NAME=$CURRENT_USER
            return 0
        else
            # This condition will probably lead to mismatched ownership of processes
            # and files requiring a chown of the GPUdb dir tree and persist dirs to fix.
            echo "Error : This install should only be run by '$USER_NAME', disallowing user '$CURRENT_USER' in '$THIS_SCRIPT'."
            echo "        Please check ownership of the directory tree of '$USER_FILE_OR_DIR', exiting."
            exit 1
        fi
    else
        if [[ "a$GPUDB_RUNNING_AS_WARN" != "a1" ]]; then
            export GPUDB_RUNNING_AS_WARN=1
            echo "Warning : USER=$USER_NAME does not exist, running as user '$CURRENT_USER' in '$THIS_SCRIPT'."
        fi
    fi

    return 0
}


# ---------------------------------------------------------------------------
# The 'sleep' program doesn't propagate CTRL-C so this script may be terminated,
# but the function calling sleep continues...  Use 'read' as a replacement, but
# only when stdin (0) is a tty, otherwise 'read' immediately fails without sleeping.

if [ -t 0 ]; then
    function sleep
    {
        local TIMEOUT=$1
        local RET_=
        # '-N 0' so keypresses don't interrupt, just CTRL-C
        # '-s' eats keystrokes, but can break the terminal after CTRL-C, user must 'reset'
        read -t "$TIMEOUT" -N 1000000 && RET_=$? || RET_=$? # work with set -e

        if [ $RET_ -eq 130 ]; then
            exit 1 # Ctrl-C caught (128+2)
        fi

        return 0 # read timeout returns 142 (man page says > 128), return success
    }
fi

# ---------------------------------------------------------------------------
# Echo, log, stdout control

# Echo a padded string in the format "Component <pad> : State", e.g. "Text : Stopped"
GPUDB_ECHO_STATUS_WIDTH=16
function echo_status
{
    local COMPONENT="$1"
    local STATE="$2"
    printf "%-${GPUDB_ECHO_STATUS_WIDTH}s : %s\n" "$COMPONENT" "$STATE"
}

# Runs the func passed in, parses out pids in the form "(1 2 3)" and echos only the numbers.
function status_to_pids
{
    local STATUS_FUNC=$1
    local S=""
    local RET=
    S=$($STATUS_FUNC) && RET=$? || RET=$? # parse even if error, some pids may be provided.
    local PIDS=$(echo "$S" | grep -oE "\([0-9 ]+\)" | sed 's/[()]//g')
    filter_docker_pids $PIDS || true
    return $RET
}

# Call this function first and then execute a program: "nohup cmd <'$LOGFILE_PIPE' &"
# When stopping call: rotatelogs_end LOGFILE
function rotatelogs_begin
{
    local LOGFILE="$1"      # /path/to/app.log
    local LOGFILE_PIPE="$2" # /path/to/app.log.pipe
    local LOG_SIZE="$3"     # 100M
    local LOG_COUNT="$4"    # 4

    local LOGFILE_LATEST=$(echo "$LOGFILE" | sed 's/\.log$//g; s/$/-latest.log/g')
    export GPUDB_ROTATELOGS_PROCESS_ID="$(hostname):$LOGFILE"

    rotatelogs_end "$LOGFILE" "$LOGFILE_PIPE" # clean-up

    if [ ! -e "$LOGFILE_PIPE" ]; then
        if ! mkfifo "$LOGFILE_PIPE"; then
          echo "ERROR: Cannot create named pipe '$LOGFILE_PIPE'"
          exit 1
        fi
    elif [ ! -p "$LOGFILE_PIPE" ]; then
        echo "ERROR: Path '$LOGFILE_PIPE' exists and is not a named pipe."
        exit 1
    fi

    touch "$LOGFILE"
    # -L creates a hard link that works when rolling
    nohup $(command -v rotatelogs) -D -f -L "$LOGFILE_LATEST" -c -n "$LOG_COUNT" "$LOGFILE" "$LOG_SIZE" >/dev/null 2>&1 <"$LOGFILE_PIPE" &
    local PID=$!

    #kill -s 0 $PID
}

function rotatelogs_end
{
    local LOGFILE="$1"      # /tmp/app.log
    local LOGFILE_PIPE="$2" # /tmp/app.log.pipe

    local LOGFILE_LATEST=$(echo "$LOGFILE" | sed 's/\.log$//g; s/$/-latest.log/g')
    export GPUDB_ROTATELOGS_PROCESS_ID="$(hostname):$LOGFILE"

    local PIDS=$(get_pids_for_app "rotatelogs .*-L $LOGFILE_LATEST" \
                                  "GPUDB_ROTATELOGS_PROCESS_ID=$GPUDB_ROTATELOGS_PROCESS_ID")
    if [ -n "$PIDS" ]; then
        kill $PIDS
    fi

    return 0
}

# ---------------------------------------------------------------------------
# Config file read/write

# Read a conf.ini parameter in the form "KEY SEPARATOR VALUE"
# OPTIONS is a single string of:
#   --match-key    : Search for exact match of "KEY SEPARATOR *", ignores sections and works on invalid files.
#                    Else use 'gpudb_conf_parser.py' in same dir as this script.
#   --warn-default : Print a warning if using the provided default value.
# Usage: get_conf_file_property [OPTIONS] VARIABLE_NAME "KEY_NAME" "=" conf.ini [OPTIONAL_DEFAULT_VALUE_IF_MISSING]
function get_conf_file_property
{
    local OPTIONS=
    while echo "$1" | grep -qE -- "$get_conf_file_property_OPTIONS"; do
        OPTIONS+="$1,"
        shift
    done

    local OUTPUT_VAR_NAME="$1"
    local KEY="$2"
    local SEPARATOR="$3"
    local FILENAME="$4"
    local DEFAULT_VALUE="${5-$get_conf_file_property_NO_DEFAULT}"

    local VALUE_GCFP_="" # unique name, no result if OUTPUT_VAR_NAME is the same

    local RET=0
    local KEYVAL_LINE=""
    if [[ "$OPTIONS" == *--match-key* ]]; then
        KEYVAL_LINE=$(grep -E "^[ $(echo -e '\t')]*${KEY}[ $(echo -e '\t')]*$SEPARATOR" "$FILENAME" | tail -n 1 | sed 's/#.*$//g; s/^[ \t]*//g; s/[ \t]*$//g' || true)
        [ -n "$KEYVAL_LINE" ] && RET=0 || RET=1
    else
        KEYVAL_LINE=$("$UTILS_THIS_SCRIPT_DIR/gpudb_conf_parser.py" "$FILENAME" "${KEY}") && RET=$? || RET=$?
    fi

    if [ $RET -ne 0 ] || [ -z "$KEYVAL_LINE" ]; then
        if [ $# -gt 4 ] && [[ "a$DEFAULT_VALUE" != "a$get_conf_file_property_NO_DEFAULT" ]]; then
            if [[ "$OPTIONS" == *--warn-default* ]]; then
                echo "WARN: Missing param in '$FILENAME', using default '$KEY' $SEPARATOR '$DEFAULT_VALUE'"
            fi
            VALUE_GCFP_="$DEFAULT_VALUE"
        else
            >&2 echo "ERROR finding conf param '$KEY' in file '$FILENAME' in '$THIS_SCRIPT', exiting."
            exit 1
        fi
    else
        VALUE_GCFP_=$(echo "$KEYVAL_LINE" | cut -d "$SEPARATOR" -f2- | sed 's/^[ \t]*//g; s/[ \t]*$//g')
    fi

    eval $OUTPUT_VAR_NAME="'$VALUE_GCFP_'"
}
# Constants for get_conf_file_property
get_conf_file_property_OPTIONS="--match-key|--warn-default"
get_conf_file_property_NO_DEFAULT="X-get_conf_file_property_NO_DEFAULT"


# Read a conf.ini parameter in the form "KEY SEPARATOR VALUE" using grep.
# Does not fail on bad files, ignores the rest of the file.
# Note that this function ignores [ini] style sections, do not prefex keys with them.
# Usage: get_conf_file_property_simple [OPTIONS] VARIABLE_NAME "KEY_NAME" "=" conf.ini [OPTIONAL_DEFAULT_VALUE_IF_MISSING]

function get_conf_file_property_simple
{
    get_conf_file_property "--match-key" "$@"
}

# Get a key='true/false' conf property, set OUTPUT_VAR_NAME to 1/0
# See get_conf_file_property() for OPTIONS.
# Usage: get_true_false_conf_file_property [OPTIONS] VARIABLE_NAME "KEY_NAME" "=" conf.ini [OPTIONAL_DEFAULT_VALUE_IF_MISSING]
function get_true_false_conf_file_property
{
    local OPTIONS=
    while echo "$1" | grep -qE -- "$get_conf_file_property_OPTIONS"; do
        OPTIONS+="$1,"
        shift
    done
    OPTIONS="${OPTIONS// /}" # remove spaces

    local OUTPUT_VAR_NAME="$1"
    local KEY="$2"
    local SEPARATOR="$3"
    local FILENAME="$4"
    local DEFAULT_VALUE="${5-$get_conf_file_property_NO_DEFAULT}"

    VALUE_GTFCFP_="" # unique name, no result if OUTPUT_VAR_NAME is the same or local
    get_conf_file_property $OPTIONS VALUE_GTFCFP_ "$KEY" "$SEPARATOR" "$FILENAME" "$DEFAULT_VALUE"

    # Using grep since "tr [:upper:] [:lower:]" gives an error when run on some
    # systems in this script, but not as the user on the command line.
    if echo "$VALUE_GTFCFP_" | grep -qi '^true$' ; then
        VALUE_GTFCFP_=1
    elif echo "$VALUE_GTFCFP_" | grep -qi '^false$' ; then
        VALUE_GTFCFP_=0
    else
        >&2 echo "Error: Unknown value in conf file: '$FILENAME' for param: '$KEY=$VALUE_GTFCFP_', expected either 'true' or 'false' in '$THIS_SCRIPT', exiting."
        exit 1
    fi

    declare -gx "$OUTPUT_VAR_NAME=$VALUE_GTFCFP_"
}

# Get a key='123' integer number conf property, set OUTPUT_VAR_NAME to value.
# See get_conf_file_property() for OPTIONS.
# Usage: get_int_conf_file_property [OPTIONS] VARIABLE_NAME "KEY_NAME" "=" conf.ini [OPTIONAL_DEFAULT_VALUE_IF_MISSING]
function get_int_conf_file_property
{
    local OPTIONS=
    while echo "$1" | grep -qE -- "$get_conf_file_property_OPTIONS"; do
        OPTIONS+="$1,"
        shift
    done
    OPTIONS="${OPTIONS// /}" # remove spaces

    local OUTPUT_VAR_NAME="$1"
    local KEY="$2"
    local SEPARATOR="$3"
    local FILENAME="$4"
    local DEFAULT_VALUE="${5-$get_conf_file_property_NO_DEFAULT}"

    VALUE_GICFP_="" # unique name, no result if OUTPUT_VAR_NAME is the same or local
    get_conf_file_property $OPTIONS VALUE_GICFP_ "$KEY" "$SEPARATOR" "$FILENAME" "$DEFAULT_VALUE"

    if ! echo "$VALUE_GICFP_" | grep -qE '^[0-9]+$' ; then
        >&2 echo "Error: Unknown value in conf file: '$FILENAME' for param: '$KEY=$VALUE_GICFP_', expected an integer in '$THIS_SCRIPT', exiting."
        exit 1
    fi

    declare -gx "$OUTPUT_VAR_NAME=$VALUE_GICFP_"
}

# Get a conf property that is a directory, handle relative paths, and verify.
# Usage: get_directory_conf_file_property VARIABLE_NAME "KEY_NAME" "=" conf.ini [OPTIONAL_DEFAULT_VALUE_IF_MISSING]
function get_directory_conf_file_property
{
    local OPTIONS=
    while echo "$1" | grep -qE -- "$get_conf_file_property_OPTIONS"; do
        OPTIONS+="$1,"
        shift
    done
    OPTIONS="${OPTIONS// /}" # remove spaces

    local OUTPUT_VAR_NAME="$1"
    local KEY="$2"
    local SEPARATOR="$3"
    local FILENAME="$4"
    local DEFAULT_VALUE="${5-$get_conf_file_property_NO_DEFAULT}"

    VALUE_GDCFP_="" # unique name, no result if OUTPUT_VAR_NAME is the same or local
    get_conf_file_property $OPTIONS VALUE_GDCFP_ "$KEY" "$SEPARATOR" "$FILENAME" "$DEFAULT_VALUE"

    PATH_GDCFP_="$VALUE_GDCFP_"
    fix_gpudb_conf_path PATH_GDCFP_ "$VALUE_GDCFP_" "$FILENAME"

    if [ ! -d "$PATH_GDCFP_" ]; then
        if [[ "a$PATH_GDCFP_" != "a$VALUE_GDCFP_" ]]; then
            echo "Error: Missing or invalid conf param value for '$KEY=$VALUE_GDCFP_' which resolves to '$PATH_GDCFP_' in '$THIS_SCRIPT', exiting."
        else
            echo "Error: Missing or invalid conf param value for '$KEY=$PATH_GDCFP_' in '$THIS_SCRIPT', exiting."
        fi

        exit 1
    fi

    declare -gx "$OUTPUT_VAR_NAME=$PATH_GDCFP_"
}

# Change a conf.ini parameter in the form "KEY = *" to "KEY = VAL" in FILENAME.
# Usage: change_conf_file_property "KEY_NAME" "=" "NEW_VALUE" conf.ini
function change_conf_file_property
{
    local KEY="$1"
    local SEPARATOR="$2"
    local NEW_VALUE="$3"
    local FILENAME="$4"

    # Also verify that it was written since sed always returns 0.
    if [[ $KEY == *"\n"* ]]; then
        eval "sed -i \"N;s@\(^[ \t]*${KEY}[ \t]*${SEPARATOR}\).*@\1${NEW_VALUE}@\" $FILENAME"
        if eval "grep -qEzo \"^[ $(echo -e '\t')]*${KEY}[ $(echo -e '\t')]*${SEPARATOR}${NEW_VALUE}\" $FILENAME" > /dev/null ; then
            echo "ERROR: Unable to update conf '$FILENAME', '$KEY'='$NEW_VALUE' in '$THIS_SCRIPT', exiting."
            exit 1
        fi
    else
        eval "sed -i \"s@\(^[ \t]*${KEY}[ \t]*${SEPARATOR}\).*@\1${NEW_VALUE}@\" $FILENAME"
        if eval "grep -qE \"^[ $(echo -e '\t')]*${KEY}[ $(echo -e '\t')]*${SEPARATOR}${NEW_VALUE}\" $FILENAME" > /dev/null; then
            echo "ERROR: Unable to update conf '$FILENAME', '$KEY'='$NEW_VALUE' in '$THIS_SCRIPT', exiting."
            exit 1
        fi
    fi
}

# Turn a relative path from the conf file into an absolute path
function fix_gpudb_conf_path
{
    local OUTPUT_VAR_NAME=$1
    local INPUT_FILE_PATH=$2
    local CONF_FILE=$3

    if [[ $INPUT_FILE_PATH != /* ]]; then
        # DO NOT call readlink in case it is a symlink and there's an error.
        # It might be more confusing that the dir has changed from the conf path.
        INPUT_FILE_PATH="$(dirname "$CONF_FILE")/$INPUT_FILE_PATH"
    fi

    declare -gx "$OUTPUT_VAR_NAME=$INPUT_FILE_PATH"
}

# ---------------------------------------------------------------------------
# File directory validation

function gpudb_check_dir
{
    if [ ! -d "$1" ]; then
        echo "Error: Missing directory: '$1' in '$THIS_SCRIPT', exiting."
        exit 1
    fi
}

function gpudb_check_file
{
    if [ ! -f "$1" ]; then
        echo "Error: Missing file: '$1' in '$THIS_SCRIPT', exiting."
        exit 1
    fi
}

# ---------------------------------------------------------------------------
# Networking

# Checks if a port is used, logs error and exits 1.
function check_port_usage
{
    local PORT="$1"
    local ERR_MSG="$2"

    if netstat -tuln | grep ":$PORT " | grep -q LISTEN; then
        echo "$ERR_MSG"
        NETSTAT_TXT=$(netstat -tulpn | grep ":$PORT " 2> /dev/null || true)
        if [[ "a$NETSTAT_TXT" != "a" ]]; then
            echo "$NETSTAT_TXT"
            ps uwwp $(echo "$NETSTAT_TXT" | awk '{print $7}' | grep -oE '[0-9]+') || true
        else
            netstat -tuln | grep "$PORT" || true
            echo "Unable to determine other app, please run 'netstat -tulpn | grep \":$PORT \"' as root to find PID"
        fi
        exit 1
    fi
}

# ---------------------------------------------------------------------------
# Pids

# Prints the given parent and child PIDs from a quoted string of parent PIDs.
function get_child_pids
{
    local PARENT_PIDS="$1" # space-separated list of pids, "1 2 3"

    local CHILD_PIDS=""
    for PID in $PARENT_PIDS; do
        local B=$(ps -ww --no-header -o pid,ppid -e | awk -vp=$PID 'function r(s){print s;s=a[s];while(s){sub(",","",s);t=s;sub(",.*","",t);sub("[0-9]+","",s);r(t)}}{a[$2]=a[$2]","$1}END{r(p)}')
        local C=$(ps -ww --no-header -o pid $B 2> /dev/null || true) # double check, doesn't print if missing.
        CHILD_PIDS+="
$C"
    done

    echo "${CHILD_PIDS}" | sort -n | uniq | tr '\n' ' ' | sed -E 's/^[ ]+//g; s/[ ]+$//g'
}

# Echos PIDs that match given 'ps auxwww' grep -E and /proc/PID/environ.
# Returns empty string and exit code 1 if no PIDs.
# Ex: get_pids_for_app "/path/to/app --config /path/to/config" "KINETICA_PROCESS_ID=$KINETICA_PROCESS_ID"
function get_pids_for_app
{
    local PS_GREP="$1"
    local ENV_GREP="$2"
    local PARENT_ONLY="${3:-}" # If 'parent' return only a single pid.

    local P=$(ps --no-header auxwww | grep -E "$PS_GREP" | grep -v 'grep -E' | awk '{print $2}')
    [ -z "$P" ] && return 1
    #>&2 echo "DEBUG get_pids_for_app '$PS_GREP' '$ENV_GREP' '$P'"

    local P2="$P"
    if [ -n "$ENV_GREP" ] && [[ "$ENV_GREP" != . ]] ; then
        local P_ENV_PATHS=$(echo "$P" | awk '{printf "/proc/"$1"/environ "}')
        # Filter out all processes that don't match given env var
        P2=$(grep -E --binary --null-data --with-filename "$ENV_GREP" $P_ENV_PATHS 2>/dev/null | tr '\0' '\n' | sed 's@^/proc/\([0-9]*\)/environ.*$@\1@g' | grep -oE '^[0-9]+$' || true)
        [ -z "$P2" ] && return 1
    fi

    P2=$(filter_docker_pids $(echo "$P2" | tr '\n' ' ') || true)
    [ -z "$P2" ] && return 1

    # Expand list to get children to avoid zombies.
    if [[ "$PARENT_ONLY" != "parent" ]]; then
        P2=$(get_child_pids "$P2" || true)
    fi
    [ -z "$P2" ] && return 1

    echo $P2 # one-line space-separated
}

function signal_pids
{
    local SIGNAL="$1"
    shift
    local PIDS="$*"

    # If no PIDs were provided, succeed
    if [ -z "$PIDS" ]; then
        return 0
    fi

    local RET=0
    kill -s "$SIGNAL" $PIDS && RET=$? || RET=$?
    if [ $RET -ne 0 ]; then
        echo "Error sending signal '$SIGNAL' to pids: $PIDS"
    fi
    return $RET
}

# Execute command until it returns success.
function wait_for_pids
{
    local PID_CMD="$1"

    # Show that we're still waiting...
    local EXIT_PROMPT_TIME_SECS=5
    local counter=0

    while ! ($PID_CMD) &> /dev/null ; do
        if [ $((counter % EXIT_PROMPT_TIME_SECS)) -eq 0 ]; then
            echo -n "."
        fi

        sleep 1
        counter=$((counter+1))
    done

    echo
}

# Kill all children of this process, ctrl-C during sleep can create orphan jobs.
# Add the 'trap' line below to your script to do process cleanup.
# trap "trap - SIGHUP SIGINT SIGTERM && kill_children" SIGHUP SIGINT SIGTERM
function kill_children
{
    local CHILD_PIDS=""

    for i in 1 2 3 4; do
        CHILD_PIDS=$(jobs -pr || true)
        if [ -z "$CHILD_PIDS" ]; then
            return 0
        fi

        if [ $i -gt 2 ]; then
            echo_status "" "$CHILD_PIDS"
            jobs -prl || true
        fi
        kill $CHILD_PIDS &> /dev/null || true
        sleep 2
    done

    CHILD_PIDS=$(jobs -pr || true)
    if [ -n "$CHILD_PIDS" ]; then
        jobs -prl
        kill -9 $CHILD_PIDS &> /dev/null || true
    fi

    return 0
}

# Returns success if currently running in docker, else return 1.
IS_RUNNING_IN_DOCKER=  # 1 if true, else 0 after calling get_is_running_in_docker
function get_is_running_in_docker
{
    if [ -z "$IS_RUNNING_IN_DOCKER" ]; then
        IS_RUNNING_IN_DOCKER=0
        if grep -qF "/docker" /proc/$$/cgroup ; then
            IS_RUNNING_IN_DOCKER=1
        fi
    fi

    [ $IS_RUNNING_IN_DOCKER -eq 1 ] && return 0 || return 1
}

# Echos the input list of pids and if running on host removes pids of docker processes.
# Returns 1 if pid list is empty
function filter_docker_pids
{
    local PIDS="$*"         # Space-separated list of pids
    local VALID_PIDS=$PIDS

    get_is_running_in_docker

    if [ -n "$PIDS" ] && ! get_is_running_in_docker; then
        VALID_PIDS=" "
        for pid in $PIDS; do
            if [ -f "/proc/${pid}/cgroup" ] && ! grep -qF "/docker" "/proc/${pid}/cgroup" > /dev/null 2>&1 ; then
                VALID_PIDS+="$pid "
            fi
        done
        VALID_PIDS=${VALID_PIDS% *}
    fi

    if [ -z "$VALID_PIDS" ]; then
        return 1
    fi

    echo $VALID_PIDS
    return 0
}

# ---------------------------------------------------------------------------

if [[ "${BASH_SOURCE[0]}" != "${0}" ]]; then
    #echo "${BASH_SOURCE[0]} is being sourced..."
    return
fi
